import numpy as np

def recalibrate(mz:list,int:list,lock_mz:float,search_tol:float,ppm_off:float=0):
    """Performs a coarse m/z recalibration based on shifting a lock mass back to target, and everything else by the same ppm shift. Applies correction
    based on the highest m/z peak within the search tolerance.
    
    :param mz: List of mz values in the spectrum to recalibrate
    :param int: Corresponding list of intensities
    :param lock_mz: Target lock mass to calibrate to - should be in the majority/all spectra
    :param search_tol: Tolerance with which to search for the lock mass (ppm)
    :param ppm_off: Optional argument specifying the previous/typical ppm error, applied if the lock mass cannot be found (default = 0, no correction)
    
    :return recalibrated_mz: Recalibrated mz if applicable, based on either the ppm error to the lock mass or optional ppm_off argument
    :return ppm_off: Applied correction in ppm"""
    ##Ignore all if lock_mz is 0 and return the original array
    if lock_mz == 0:
        return mz, int
    else:
        ##Compute ppm error of every m/z in the spectrum to the lock mass     
        diff_mz_ppm = (mz - lock_mz)/lock_mz * 1e6

        ##Find candidate m/z within the tolerance window
        iter=-1
        candidate_mz =[]
        candidate_int = []
        for ppm_diff in diff_mz_ppm:
            iter += 1
            if abs(ppm_diff) <= search_tol:
                candidate_mz.append(mz[iter])
                candidate_int.append(int[iter])
        
        ##take the highest intensity peak in the tolerance window as the lock mass, and identify the correction ppm error based on it
        try:
            match_idx = np.where(candidate_int == max(candidate_int))[0][0]
            id_mz = candidate_mz[match_idx]
            ppm_off = (id_mz - lock_mz)/lock_mz * 1e6
        except:
            pass

        ##Compute recalibrated mass spectrum based on identified lock mass. If above failed (lock mass not identified within tolerance window)
        ##Than the spectrum is corrected based on input 'ppm_off' variable, typically whatever the correction was on the previous spectrum
        recalibrated_mz = mz - (ppm_off * mz / 1e6)
        return recalibrated_mz, ppm_off




    



import scipy.ndimage
import cv2 as cv
import numpy as np
from tkinter import ttk, filedialog, messagebox
import tkinter as tk
import matplotlib.pyplot as plt
from matplotlib.widgets import PolygonSelector
from matplotlib.path import Path
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2Tk
from imzml_writer import utils


roi_mask = None
verts = None
mask_list = None
TEAL = "#2da7ad"
BEIGE = "#dbc076"
FONT = ("HELVETICA", 18, 'bold')


def unsharp_mask(image, kernel_size=(5, 5), sigmaX=1.0, sigmaY=1.0, amount=1.0, threshold=0):
    """Return a sharpened version of the image, using an unsharp mask."""
    blurred = cv.GaussianBlur(image, kernel_size, sigmaX=sigmaX, sigmaY=sigmaY)
    sharpened = float(amount + 1) * image - float(amount) * blurred
    sharpened = np.maximum(sharpened, np.zeros(sharpened.shape))
    sharpened = np.minimum(sharpened, np.max(image) * np.ones(sharpened.shape))
    # sharpened = sharpened.round().astype(np.uint8)
    if threshold > 0:
        low_contrast_mask = np.absolute(image - blurred) < threshold
        np.copyto(sharpened, image, where=low_contrast_mask)
    return sharpened

def smooth_image(img_data,asp:float, factor:int=3,base_sigma:float=10,weight_factor:float=0.5):
    zoomed_img = scipy.ndimage.zoom(img_data,factor)
    sharpened_img = unsharp_mask(zoomed_img, sigmaX=base_sigma, sigmaY=base_sigma/asp, kernel_size=(9,9), amount=weight_factor)
    return sharpened_img

def ROI_select(ROI_button:tk.Button, ion_image:np.array, aspect_ratio:float, color_NL=float):
    global mask_list
    roi_selector = tk.Toplevel()
    roi_selector.title("ROI Selector")
    roi_selector.config(padx=5,pady=5,bg=TEAL)

    fig, ax = plt.subplots()
    canvas = FigureCanvasTkAgg(fig, master=roi_selector)
    plt.imshow(ion_image,aspect=aspect_ratio,interpolation="none",vmin=0,vmax=color_NL,cmap='viridis')
    plt.title("Select your ROI")
    
    selector = []
    selector.append(PolygonSelector(ax, props=dict(color='red')))
    fig.set_facecolor(TEAL)
    fig.tight_layout()
    canvas.get_tk_widget().pack()

    def add_additional_ROI():
        nonlocal selector
        selector.append(PolygonSelector(ax, props=dict(color='red')))

    add_another = tk.Button(roi_selector, text="Add another inclusion region", bg=TEAL,highlightbackground=TEAL,command=add_additional_ROI)
    add_another.pack()

    def quit_select():
        nonlocal roi_selector
        mask_list = []
        vert_list = []
        for idx, roi in enumerate(selector):
            if len(roi.verts) > 2:
                # Create coordinate arrays
                h, w = ion_image.shape[:2]
                y, x = np.mgrid[:h, :w]
                coors = np.hstack((x.reshape(-1, 1), y.reshape(-1, 1)))  # (N, 2) array
                
                # Create path from vertices
                path = Path(roi.verts)
                mask = path.contains_points(coors)
                mask = mask.reshape(h, w).astype(int)  # Reshape to image dimensions
                mask_list.append(mask)
                vert_list.append(roi.verts)


        if len(mask_list) >= 1:
            global roi_mask, verts
            final_mask = mask_list[0]
            for mask in mask_list:
                final_mask = np.maximum(final_mask,mask)
            roi_mask = final_mask

            all_verts = []
            for idx, vert in enumerate(vert_list):
                for vertex in vert:
                    all_verts.append((idx, vertex[0], vertex[1]))
            
            verts = all_verts
        plt.close(fig)
        canvas.get_tk_widget().destroy()
        roi_selector.destroy()

    roi_selector.protocol("WM_DELETE_WINDOW", quit_select)
    quit_selecting = tk.Button(roi_selector, text="Stop selecting", bg=TEAL,highlightbackground=TEAL, command=quit_select)
    quit_selecting.pack()



def save_ROI():
    save_location = filedialog.asksaveasfile(defaultextension='.npz')
    np.savez(save_location.name, roi_mask=roi_mask, verts=verts, mask_list=mask_list)

def load_ROI():
    global roi_mask, verts
    file_location =filedialog.askopenfilename(defaultextension='.npz')

    loaded_data = np.load(file_location,allow_pickle=True)
    verts = loaded_data['verts']
    roi_mask = loaded_data['roi_mask']
    mask_list = loaded_data['mask_list']

def write_masked_imzml_handler(tgt_file:str):

    #Verify roi mask actually exists
    if roi_mask is None:
        messagebox.showwarning(title='No mask imposed',message="Please specify an ROI mask before trying to export")
        return

    progress_window = tk.Toplevel()
    progress_window.title("Writing masked imzML...")
    progress_window.config(padx=20, pady=10, bg=TEAL)
    ttk.Label(progress_window, text="Writing masked imzML please wait...", background=TEAL, foreground="white", font=FONT).pack()
    progress_window.update_idletasks()


    try:
        filepath = utils.write_masked_imzML(tgt_file,roi_mask)
        messagebox.showinfo(title="imzML write complete", message=f"Masked file saved to {filepath}")
    except Exception as exc:
        messagebox.showerror(title="Write failed", message=str(exc))
    finally:
        progress_window.grab_release()
        progress_window.destroy()

    

    





        


        



import docker
import subprocess
import shutil
import os
import sys
import pymzml
import numpy as np
import pyimzml.ImzMLWriter as imzmlw
from bs4 import BeautifulSoup, Tag
import string
import re
import tkinter as tk
from tkinter import filedialog,messagebox
import time
import json
import logging
import warnings
import pyimzml.ImzMLParser as imzmlp
from pathlib import Path

from imzml_writer.recalibrate_mz import recalibrate
from imzml_writer import __version__

logger = logging.getLogger(__name__)
logging.basicConfig(filename = "LOGFILE_imzML_Writer.log",level = logging.INFO)
logger.info("NEW INSTANCE STARTING...")


##Colors and FONTS
TEAL = "#2da7ad"
BEIGE = "#dbc076"
GREEN = "#22d10f"
FONT = ("HELVETICA", 18, 'bold')

def get_drives():
    """On windows machines, retrieves the accessible drives (e.g C:\\, D:\\, etc.) in to for automated seeking
    of msconvert.

    :return: Available drives, as a list of strings."""
    from ctypes import windll
    drives = []
    bitmask = windll.kernel32.GetLogicalDrives()
    for letter in string.ascii_uppercase:
        if bitmask & 1:
            drives.append(letter)
        
        bitmask >>= 1
    
    return drives


def find_file(target:str, folder:str):
    """Recursely searches the folder for the target file - helps find msconvert in cases
    where it isn't specified in the path.
    
    :param target: Target file as a string
    :param folder: Top-level folder to search through
    :return: full path to file if found, [ ] if not present"""
    try:
        for f in os.listdir(folder):
            path = os.path.join(folder,f)
            if os.path.isdir(path):
                result = find_file(target, path)
                if result is not None:
                    return result
                continue
            if f == target:
                return path
    except Exception as e:
        pass

def msconvert_searchUI() -> str:
    """Launches a dialog window to ask the user whether to search manually or automatically for msconvert install path
    
    :return: Specified mode to search for msconvert ("auto" or "manual")"""
    search_mode = ""
    def set_auto():
        nonlocal search_mode
        search_mode = "auto"
        msconv_finder.destroy()
    def set_manual():
        nonlocal search_mode
        search_mode = "manual"
        msconv_finder.destroy()

    msconv_finder = tk.Tk()
    msconv_finder.title("Find msconvert...")
    msconv_finder.config(padx=5,pady=5,bg=TEAL)

    autofind = tk.Button(msconv_finder,text="Auto Search...(Warning : Slow for large drives)",bg=TEAL,highlightbackground=TEAL,command=set_auto)
    autofind.grid(row=1,column=1,padx=15,pady=15)

    man_find = tk.Button(msconv_finder,text="Manual Search",bg=TEAL,highlightbackground=TEAL,command=set_manual)
    man_find.grid(row=1,column=2,padx=15,pady=15)

    while search_mode == "":
        msconv_finder.update()
        time.sleep(0.5)

    return search_mode

def autofind_msconvert():
    """Finds msconvert by searching all available drives, verifies success by calling
    info of msconvert
    
    :return: Full path to msconvert.exe"""
    drives = get_drives()

    candidates = []
    for drive in drives:
        drive_str = (f"{drive}:\\".__repr__()).replace("'","")
        candidates.append(find_file("msconvert.exe",drive_str))
    

    for candidate in candidates:
        if candidate is not None:
            if "msconvert.exe" in candidate:
                res = subprocess.run(candidate, shell=True,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.STDOUT,
                            stdin=subprocess.PIPE,
                            cwd=os.getcwd(),
                            env=os.environ)
                if res.returncode == 0:
                    #msconvert successfully found and called
                    return candidate


def check_msconvert():
    """Checks that msconvert is available for the current python environment - returns msconvert path/callable"""
    msconvert = "msconvert"
    try:
        res = subprocess.run(msconvert, shell=True,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.STDOUT,
                            stdin=subprocess.PIPE,
                            cwd=os.getcwd(),
                            env=os.environ) 
        if res.returncode == 0:
            # Found as default
            return msconvert
            
        elif res.returncode != 0:
            try:
                mod_path = os.path.dirname(os.path.abspath(__file__))
                settings_path = os.path.join(mod_path,"msconvert_path.json")
                with open(settings_path,'r') as file:
                    path_data = json.load(file)
                msconvert = path_data["msconvert_path"]
                res = subprocess.run(msconvert, shell=True,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.STDOUT,
                            stdin=subprocess.PIPE,
                            cwd=os.getcwd(),
                            env=os.environ)
                
                if res.returncode != 0:
                    raise
                # Found in saved config
                return msconvert
            except:
                    search_method = msconvert_searchUI()
                    if search_method == "manual":
                        msconvert = filedialog.askopenfilename(initialdir=os.getcwd(),title="Please select msconvert.exe",filetypes=[("msconvert.exe","msconvert.exe")])
                        msconvert = os.path.abspath(msconvert)
                    elif search_method == "auto":
                        msconvert = autofind_msconvert()
                        msconvert = os.path.abspath(msconvert)
                    res = subprocess.run(msconvert, shell=True,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.STDOUT,
                        stdin=subprocess.PIPE,
                        cwd=os.getcwd(),
                        env=os.environ)
                    if res.returncode == 0:
                        # Found by manual/autosearch, saving to config
                        mod_path = os.path.dirname(os.path.abspath(__file__))
                        settings_path = os.path.join(mod_path,"msconvert_path.json")
                        set_path = {"msconvert_path": msconvert}
                        with open(settings_path,'w') as file:
                            json.dump(set_path,file)
                        
                        return msconvert
                    else:
                        raise
    except:
        raise Exception("msConvert not available, check installation and verify msConvert path is specified correctly")


def viaPWIZ(path:str,write_mode:str,combine_ion_mobility:bool):
    """Method to call msconvert directly if the detected platform is on windows. Converts all target files in the path to mzML in the specified mode.

    :param path: path to the target files
    :param write_mode: "Centroid" or "Profile" modes
    :param combine_ion_mobility: Whether or not --combineIonMobility flag is passed to msconvert
    :return: None"""
    ##check pwiz availability:
    file_type = get_file_type(path)
    current_dir = os.getcwd()
    os.chdir(path)

    
    msconvert = check_msconvert()

    ms_convert_args = [msconvert,fr"{path}\*.{file_type}", "--mzML", "--64"]
    if write_mode=="Centroid":
        ms_convert_args.append("--filter")
        ms_convert_args.append("peakPicking true 1-")
    
    ms_convert_args.append("--simAsSpectra")
    ms_convert_args.append("--srmAsSpectra")
    if combine_ion_mobility:
        ms_convert_args.append("--combineIonMobility")
    


    convert_process = subprocess.run(ms_convert_args,
                                       stdout=subprocess.DEVNULL,
                                       shell=True,
                                       stderr=subprocess.STDOUT,
                                       stdin=subprocess.PIPE,
                                       cwd=os.getcwd(),
                                       env=os.environ)


    os.chdir(current_dir)

##The below three functions (tryint, alphanum_keys, human_sort) are borrowed from an excellent post by Ned Batchelder for 'natural sorting' in Python
##http://nedbatchelder.com/blog/200712/human_sorting.html
def tryint(s):
    """
    Part of the human sorting collection of functions borrowed from http://nedbatchelder.com/blog/200712/human_sorting.html. Returns an int if possible, or `s` unchanged.

    :param s: Trial variable to test if it can be converted to an integer
    :return: integer if convertible, s if not.
    """
    try:
        return int(s)
    except ValueError:
        return s

def alphanum_key(s:str) -> list:
    """
    Part of the human sortable collection of functions borrowed from http://nedbatchelder.com/blog/200712/human_sorting.html. Turn a string into a list of string and number chunks.

    :param s: String to be chunked out
    :return: List of string/number chunks
    """
    return [ tryint(c) for c in re.split('([0-9]+)', s) ]

def human_sort(l:list) -> list:
    """
    Part of the human sortable collection of functions borrowed from http://nedbatchelder.com/blog/200712/human_sorting.html. Sorts a list in the way that humans expect.

    :param l: List to be sorted in a human-intuitive wave.
    :return: Sorted list.
    """
    return l.sort(key=alphanum_key)


def get_file_type(path:str):
    """Identifies the most abundant file type in the specified path, ignoring hidden files.

    :param path: path to files specified as a string.
    :return: Most abundant file extension in path"""
    files = os.listdir(path)
    file_poss = {}
    for file in files:
        if not file.startswith("."):
            ext = file.split(".")[-1]
            if ext not in file_poss.keys():
                file_poss[ext] = 1
            else:
                file_poss[ext]+= 1
    
    num_biggest = 0
    for ext in file_poss.keys():
        if file_poss[ext] > num_biggest:
            extension = ext
            num_biggest = file_poss[ext]
    
    return extension

def Check_Docker_Image():
    """Tests that docker is available, prompts the user to update/install if available"""
    DOCKER_IMAGE = "chambm/pwiz-skyline-i-agree-to-the-vendor-licenses"
    try:
        client = docker.from_env()
    except:
        res = subprocess.run(["open", "--background", "-a", "Docker"])
        if res.returncode == 0:
            time.sleep(2.5)
            client = docker.from_env()
        else:
            messagebox.showwarning(title="No Docker",message="Docker unavailable - please launch/install Docker desktop before proceeding...")
            client = docker.from_env()

    try:
        data = client.images.get(DOCKER_IMAGE)
        if "latest" not in str(data):
            resp = messagebox.askquestion("Newer docker image available", "A newer version of the msconvert docker image is available, would you like to update?")
            if resp == "yes":
                client.images.pull(DOCKER_IMAGE)
    except:
        resp = messagebox.askquestion("Docker image unavailable", "No docker image for msconvert is available, would you like to download it now? (WARNING: May take several minutes)")
        if resp == "yes":
            client.images.pull(DOCKER_IMAGE)
        else:
            raise
    
    return client



def RAW_to_mzML(path:str,write_mode:str="Centroid", combine_ion_mobility:bool=False):
    """Calls msConvert via docker on linux and Mac, or calls viaPwiz method on PC to manage conversion of raw vendor files to mzML format within the specified path

    :param path: path to files containing raw instrument data.
    :param write_mode: Write mode for msconvert - 'Profile' or 'Centroid'.
    :param combine_ion_mobility: Whether or not to hand --combineIonMobilitySpectra flag to msconvert (default false)"""
    if "win" in sys.platform and sys.platform != "darwin":
        viaPWIZ(path,write_mode, combine_ion_mobility)
    else:
        DOCKER_IMAGE = "chambm/pwiz-skyline-i-agree-to-the-vendor-licenses"
        client = Check_Docker_Image()
        file_type = get_file_type(path)

        vol = {path: {'bind': fr"/{DOCKER_IMAGE}/data", 'mode': 'rw'}}


        comm = fr"wine msconvert /{DOCKER_IMAGE}/data/*.{file_type} --zlib=off --mzML --64 --outdir /{DOCKER_IMAGE}/data"

        if write_mode=="Centroid":
            comm += fr" --filter '"'peakPicking true 1-'"'"

        comm+= " --simAsSpectra --srmAsSpectra"
        if combine_ion_mobility:
            comm += " --combineIonMobilitySpectra"

        env_vars = {"WINEDEBUG": "-all"}
        
        ##Call/run the docker container
        client.containers.run(
            image=DOCKER_IMAGE,
            environment=env_vars,
            volumes = vol,
            command=comm,
            working_dir=path,
            auto_remove= True,
            detach=False
            )
        
        

def clean_raw_files(path:str,file_type:str):
    """Cleans up file system after RAW_to_mzML has completed, creating two folders within the specified path:

    **Initial RAW files** - raw vendor files

    **Output mzML Files** - processed mzML files output by msConvert

    :param path: path to directory to clean up
    :param file_type: extension for raw vendor data to place into raw file directory"""
    mzML_folder = os.path.join(path,"Output mzML Files")
    RAW_folder = os.path.join(path,"Initial RAW files")
    if not os.path.isdir(mzML_folder):
        os.mkdir(mzML_folder)
    if not os.path.isdir(RAW_folder):
        os.mkdir(RAW_folder)

    for file in os.listdir(path):
        if not file.startswith("."):
            logger.info(f"Starting file transfer: {file}")
            if ".mzML".lower() in file.lower():
                shutil.move(os.path.join(path,file),os.path.join(mzML_folder,file))
            elif file_type in file and file != "Initial RAW files":
                shutil.move(os.path.join(path,file),os.path.join(RAW_folder,file))

def get_final_scan_time(run:pymzml.run.Reader):
    """Returns the final scan time from the specified mzML

    :param run: pymzml reader object
    :return scan_time: Scan time in minutes (float)"""
    for spec in run:
        scan_time = spec.scan_time_in_minutes()
    
    return scan_time

def mzML_to_imzML_convert(progress_target=None,PATH:str=os.getcwd(),LOCK_MASS:float=0,TOLERANCE:float=20,zero_indexed:bool=False,no_duplicating:bool=False,scan_mode:str = "x-scan"):
    """Handles conversion of mzML files to the imzML format using the pyimzml library. Converts data line-by-line (one mzML at a time),
    aligning data based on scan time and splitting into separate imzML files for each scan in the source mzML.
    
    :param progress_target: tkinter progress bar object from the GUI to update as conversion progresses
    :param PATH: - Working path for source mzML files
    :param LOCK_MASS: - m/z to use for coarse m/z recalibration if desired. 0 = No recalibration
    :param TOLERANCE: Search tolerance (in ppm) with which to correct m/z based on the specified lock mass. Default 20 ppm
    :param zero_indexed: Specifies whether pixel dimensions should start from 1 (default - False) or 0 (True)
    :param no_duplicating: Specifies whether spectra can be duplicated into adjacent pixels for sparsely sampled lines. Default True
    :param scan_mode: Whether the data was acquired in 'x-scan' or 'y-scan' mode."""

    ##Ensure lock mass and tolerance are formatted as float
    LOCK_MASS = float(LOCK_MASS)
    TOLERANCE = float(TOLERANCE)
    files = os.listdir(PATH)
    human_sort(files)

    ##Extracts filter strings, num pixels for each scan, etc
    scan_filts=[]
    polarities = []
    ms_levels = []
    file_iter=-1
    spectrum_counts=[]
    mzml_files=[]
    spec_counts=[]
    list_type = False
    for file in files:
        if ".mzML".lower() in file.lower():
            file_iter+=1
            tmp = pymzml.run.Reader(os.path.join(PATH,file))
            spec_counts.append(tmp.get_spectrum_count())
            ##Ignore partially collected datafiles that were cut-short (threshold of <50% the scans of the mean datafile)
            if np.mean(spec_counts)*0.5 > tmp.get_spectrum_count():
                break
            
            
            mzml_files.append(file)
            

            ##Retrieve list of filter strings from first file
            if file_iter==0:
                for spectrum in tmp:
                    if isinstance(spectrum["filter string"],list):
                            list_type = True
                    if list_type:
                        logger.warning(f"LIST TYPE filter strings detected")
                        scan_filts = spectrum["filter string"][0]
                        ms_levels = int(spectrum['MS:1000511'])
                        if spectrum["MS:1000129"]:
                            polarities = "negative"
                        elif spectrum["MS:1000130"]:
                            polarities = "positive"
                    else: 
                        if spectrum["filter string"] not in scan_filts:
                            scan_filts.append(spectrum["filter string"])
                            ms_levels.append(int(spectrum['MS:1000511']))
                            if spectrum["MS:1000129"]:
                                polarities.append("negative")
                            elif spectrum["MS:1000130"]:
                                polarities.append("positive")

            if not list_type:
                tmp_spectrum_counts = {filt_name:0 for filt_name in scan_filts}
            else:
                str_list = set(spectrum["filter string"])
                str_list = list(str_list)
                tmp_spectrum_counts={}
                for entry in str_list:
                    tmp_spectrum_counts[entry]=0

            for spectrum in tmp:
                if not list_type:
                    tmp_spectrum_counts[spectrum["filter string"]] += 1
                elif list_type:
                    tmp_spectrum_counts[spectrum["filter string"][0]]+= 1
            
            spectrum_counts.append(tmp_spectrum_counts)

    tmp.close()
    del spectrum

    #Find conserved portion of name for output filename
    str_array = [letter for letter in mzml_files[0]]
    OUTPUT_NAME = "".join(str_array)
    while OUTPUT_NAME not in mzml_files[-1]:
        str_array.pop(-1)
        OUTPUT_NAME = "".join(str_array)

    #Compute max number of pixels in each scan filter to construct pixel grids
    max_x_pixels = {}
    y_pixels = len(spectrum_counts)
    contender_idx=[]
    if isinstance(scan_filts,str):
        scan_filts = [scan_filts]

    for filt in scan_filts:
        max_x = 0
        idx =-1
        for spec_file in spectrum_counts:
            idx += 1
            if spec_file[filt] > max_x:
                max_x = spec_file[filt]
                contender_idx.append(idx)
        max_x_pixels[filt] = max_x

    #Retrieve max times for time-alignment on longest spectra, build ideal time array
    max_times = []
    for idx in contender_idx:
        tmp = pymzml.run.Reader(os.path.join(PATH,mzml_files[idx]))
        scan_time = get_final_scan_time(tmp)
        max_times.append(scan_time)

    time_targets={}
    iter = -1
    for key in max_x_pixels:
        iter += 1
        time_array = np.linspace(0,max_times[iter],max_x_pixels[key])
        time_targets[key] = time_array

    #Initiate imzmL objects
    image_files = {}
    output_files ={}
    for filt_idx, filt in enumerate(scan_filts):
        if filt == None:
            image_files[filt]=imzmlw.ImzMLWriter(output_filename=fr"{OUTPUT_NAME}_None",mode="processed",polarity=polarities[filt_idx])
        else:
            image_files[filt] = imzmlw.ImzMLWriter(output_filename=fr"{OUTPUT_NAME}_{filt.split(".")[0]}",mode="processed",polarity=polarities[filt_idx])
        output_files[filt]=(fr"{OUTPUT_NAME}_{filt}")

    num_duplicates = 0
    num_total = 0
    #Build image grid, write directly to an imzML
    for y_row in range(y_pixels):
        active_file = pymzml.run.Reader(os.path.join(PATH,mzml_files[y_row]))
        for filt in scan_filts:
            tmp_times = []
            spec_list = []
            for spectrum in active_file:
                if spectrum["filter string"] == filt:
                    tmp_times.append(spectrum.scan_time_in_minutes())
                    spec_list.append(spectrum)
                elif list_type:
                    tmp_times.append(spectrum.scan_time_in_minutes())
                    spec_list.append(spectrum)

            pvs_ppm_off = 0
            used_idx = []
            
            for x_row in range(max_x_pixels[filt]):
                align_time = time_targets[filt][x_row]
                time_diffs = abs(tmp_times - align_time)
                match_idx = np.where(time_diffs == min(time_diffs))[0][0]
                num_total += 1
                if zero_indexed:
                    x_coord = x_row
                    y_coord = y_row
                else:
                    x_coord = x_row + 1
                    y_coord = y_row + 1
                
                if scan_mode == "x-scan":
                    coords = (x_coord, y_coord, 1)
                elif scan_mode == 'y-scan':
                    coords = (y_coord, x_coord, 1)
                
                if not match_idx in used_idx:
                    used_idx.append(match_idx)
                    if len(used_idx) > 15:
                        used_idx.pop(0)

                    match_spectra = spec_list[match_idx]
                    [recalibrated_mz, pvs_ppm_off] = recalibrate(mz=match_spectra.mz, int=match_spectra.i,lock_mz=LOCK_MASS,search_tol=TOLERANCE,ppm_off=pvs_ppm_off)
                    if len(recalibrated_mz) != 0:
                        image_files[filt].addSpectrum(recalibrated_mz,match_spectra.i,coords)
                else:
                    num_duplicates += 1
                    if not no_duplicating:
                        match_spectra = spec_list[match_idx]
                        [recalibrated_mz, pvs_ppm_off] = recalibrate(mz=match_spectra.mz, int=match_spectra.i,lock_mz=LOCK_MASS,search_tol=TOLERANCE,ppm_off=pvs_ppm_off)
                        if len(recalibrated_mz) != 0:
                            image_files[filt].addSpectrum(recalibrated_mz,match_spectra.i,coords,userParams=[{"name":"DuplicatedSpectrum","value":"True"}])

                    # print(f"Duplicated pixel detected - {num_duplicates} / {num_total} ({num_duplicates * 100 / num_total:.2f}%)")


        ##Update progress bar in the GUI as each mzML finishes
        progress = int(y_row*100/(y_pixels-1)) 

        if progress > 0 and progress_target != None:   
            progress_target.stop() 
            progress_target.config(mode="determinate",value=int(y_row*100/(y_pixels-1)))

    ##Close imzML objects for downstream annotation
    update_files = os.listdir()
    update_files.sort()

    for filt in scan_filts:
        image_files[filt].close()

def imzML_metadata_process(model_files:str,x_speed:float,y_step:float,path:str,tgt_progress=None, scan_mode:str = "x-scan"):
    """Manages annotation of imzML files with metadata from source mzML files and user-specified fields (GUI). 
    
    :param model_files: Directory to the folder containing mzML files
    :param x_speed: scan speed in the x-direction, µm/sec
    :param y_step: step between strip lines, µm
    :param path: path to the directory where imzML files should be stored after annotation
    :param tgt_progress: Tkinter progress bar object to update as the process continues
    :param scan_mode: Whether the data was acquired in 'x-scan' or 'y-scan' mode."""
    
    global OUTPUT_NAME, time_targets

    ##Retrieve and sort files from the working directory (imzML) and model file directory (mzML)
    update_files = os.listdir()
    update_files.sort()

    scan_filts=[]
    polarities = []
    ms_levels = []
    model_file_list = os.listdir(model_files)
    model_file_list.sort()

    ##Ignore hidden files
    while model_file_list[0].startswith("."):
        model_file_list.pop(0)

    ##Extract filter strings from the first mzML source file
    tmp = pymzml.run.Reader(os.path.join(model_files,model_file_list[0]))
    for spectrum in tmp:
        if spectrum["filter string"] not in scan_filts:
            scan_filts.append(spectrum["filter string"])
            ms_levels.append(int(spectrum['MS:1000511'])) #record ms_level
            if spectrum["MS:1000129"]: #record polarities
                polarities.append("negative") 
            elif spectrum["MS:1000130"]:
                polarities.append("positive")
    
        # final_time_point = spectrum["scan time"]
        final_time_point = spectrum.scan_time_in_minutes()

    ##Extract common output name based on common characters in first and last mzML file
    str_array = [letter for letter in model_file_list[0]]
    OUTPUT_NAME = "".join(str_array)
    while OUTPUT_NAME not in model_file_list[-1]:
        str_array.pop(-1)
        OUTPUT_NAME = "".join(str_array)


    ##Loop to annotate each imzML file
    iter = 0
    for filt_idx, filt in enumerate(scan_filts):
        #Find the target file based on a filter string match
        iter+=1
        for file in update_files:
            if ".imzML" in file:
                partial_filter_string = file.split(OUTPUT_NAME+"_")[-1].split(".imzML")[0]
                if partial_filter_string == "None":
                    target_file = file
                elif partial_filter_string in filt:
                    target_file = file

        ##Calls the actual annotation function
        annotate_imzML(
                    annotate_file=target_file,
                    SRC_mzML=os.path.join(model_files,model_file_list[0]),
                    scan_time=final_time_point,
                    filter_string=filt,
                    x_speed=x_speed,
                    y_step=y_step,
                    ms_level = ms_levels[filt_idx],
                    polarity = polarities[filt_idx],
                    scan_mode = scan_mode)

        ##Update progress bar in the GUI
        progress = int(iter*100/len(scan_filts))
        if progress > 0 and tgt_progress != None:
            tgt_progress.stop()
            tgt_progress.config(mode="determinate",value=progress)

    ##After conversion is complete, clean up files by putting the annotated imzML files in a new directory within the datafile folder           
    move_files(OUTPUT_NAME,path)

def move_files(probe_txt:str,path:str):
    """Moves files matching a search string (probe_txt) in the current working directory into the specified directory in a new folder called 'probe_txt'
    
    :param probe_txt: The search string to find in the current directory.
    :param path: The target directory to move files to"""
    files = os.listdir()
    new_directory = os.path.join(path, probe_txt)
    try:
        print(new_directory)
        os.makedirs(new_directory, exist_ok=True)
    except:
        pass
    
    for file in files:
        if probe_txt in file:
                try:
                    shutil.copy2(file, os.path.join(new_directory,file))
                except Exception as e:
                    logger.info("Copy object claimed failure")
                finally:
                    os.remove(file)


def annotate_imzML(annotate_file:str,SRC_mzML:str,scan_time:float=0.001,filter_string:str="none given",x_speed:float=1,y_step:float=1,polarity:str="positive",ms_level:int=1, scan_mode:str = 'x-scan'):
    """Takes pyimzml output imzML files and annotates them using GUI inputs and the corresponding mzML source file, then cleans up errors in the imzML structure
    for compatibility with imzML viewers/processors.

    :param annotate_file: the imzML file to be annotated
    :param SRC_mzML: the source file to pull metadata from
    :param scan_time: The total time required to scan across the imaging area at speed x_speed (mins)
    :param filter_string: what scan filter is actually captured  (default = "none given")
    :param x_speed: The scan speed across the imaging area during linescans (µm/s)
    :param y_step: The distance between adjacent strip lines across the imaging area (µm/s)
    :param scan_mode: Whether the data was acquired in 'x-scan' or 'y-scan' mode.
    """

    #Error handling for when scan filter extraction fails
    result_file = annotate_file
    if filter_string == None:
        filter_string = "None"

    #Retrieve data from source mzml
    with open(SRC_mzML) as file:
        data = file.read()
    data = BeautifulSoup(data,'xml')

    #Grab instrument model from the source mzML
    try:
        instrument_model = data.referenceableParamGroup.cvParam.get("name")
    except:
        instrument_model = "Could not find"

    #Open un-annotated imzML
    with open(annotate_file) as file:
        data_need_annotation = file.read()
    data_need_annotation = BeautifulSoup(data_need_annotation,'xml')

    #Replace template data with key metadata from mzML
    replace_list = ['instrumentConfigurationList']
    for replace_item in replace_list:
        data_need_annotation.find(replace_item).replace_with(data.find(replace_item))

    #Write instrument model to imzML, filter string
    data_need_annotation.instrumentConfigurationList.instrumentConfiguration.attrs['id']=instrument_model
    new_tag = Tag(builder=data_need_annotation.builder,
                  name="cvParam",
                  attrs={'accession':'MS:1000031',"cvRef":"MS","name":instrument_model})
    # new_tag = data_need_annotation.new_tag("cvParam", accession="MS:1000031", cvRef="MS")
    data_need_annotation.instrumentConfigurationList.instrumentConfiguration.append(new_tag)

    #Remove empty instrument ref from imzML template
    for paramgroup in data_need_annotation.select("referenceableParamGroupRef"):
        if paramgroup['ref']=="CommonInstrumentParams":
            paramgroup.extract()
    
    for cvParam in data_need_annotation.select("cvParam"):
        if cvParam["accession"]=="MS:1000530":
            del cvParam["value"]
        if cvParam["accession"]=="IMS:1000411":
            cvParam["accession"]="IMS:1000413"
            cvParam["name"]="flyback"
        ##Future y-scan mode - accession:
#         [Term]
# id: IMS:1000481
# name: vertical line scan
# def: "The scanning line is a vertical one." [COMPUTIS:IMS]
# is_a: IMS:1000048 ! Scan Type



    for tag in data_need_annotation.referenceableParamGroupList:
        if "scan1" in str(tag):
            for tag2 in tag:
                if "MS:1000512" in str(tag2):
                    tag2["value"] = filter_string
                    

        
    #Read pixel grid information from imzML
    for tag in data_need_annotation.scanSettingsList.scanSettings:
        if 'cvParam' in str(tag):
            if tag.get("accession") == "IMS:1000042": #num pixels x
                x_pixels = tag.get("value")
            elif tag.get("accession") == "IMS:1000043": #num pixels y
                y_pixels = tag.get("value")

    #Calculate pixel sizes and overall dimensions from size of pixel grid, scan speed, step sizes
    if scan_mode == "x-scan":
        x_pix_size = float(x_speed * scan_time * 60 / float(x_pixels))
        y_pix_size = y_step
    elif scan_mode == "y-scan":
        y_pix_size = float(x_speed * scan_time * 60 / float(y_pixels))
        x_pix_size = y_step
    
    max_x = int(x_pix_size * float(x_pixels))
    max_y = int(y_pix_size * float(y_pixels))



    ##TODO - Test changing this into 'pixel size x' instead for compatibility with all, if everyone still works will leave as default. Otherwise another advanced tab for mutually exclusive compatibility?
    accessions = ["IMS:1000046", "IMS:1000047", "IMS:1000044", "IMS:1000045"]
    names = ["pixel size x", "pixel size y", "max dimension x", "max dimension y"]
    values = [x_pix_size, y_pix_size, max_x, max_y]

    #Actual insertion of data - need to write string into a beautiful soup object with NO FORMATTING to append it
    for i in range(4):
        append_item = f'<cvParam cvRef="IMS" accession="{accessions[i]}" name="{names[i]}" value="{values[i]}"/>\n'
        append_item = BeautifulSoup(append_item,'xml')
        data_need_annotation.scanSettingsList.scanSettings.append(append_item)


    for cvParam in data_need_annotation.select("cvParam"):
        if cvParam["accession"] in accessions:
            cvParam["unitCvRef"]="UO"
            cvParam["unitAccession"]="UO:0000017"
            cvParam["unitName"]="micrometer"

                    
    ##Specify imzML writer involvement and version in the resulting imzML
    for soft_list in data_need_annotation.select("softwareList"):
        count = int(soft_list.attrs['count']) + 1
        new_tag = Tag(builder=data_need_annotation.builder,
                    name = "software",
                    attrs = {'id':'imzML_Writer',"version":__version__})
        descr_tag = Tag(builder=data_need_annotation.builder,
                    name = "cvParam",
                    attrs = {'accession':'MS:1000799',"cvRef":"MS","name":"Custom unreleased software tool","value":f"imzML Writer v{__version__}"})
        

        soft_list.append(new_tag)
        for software in soft_list.select("software"):
            if software.attrs["id"] == "imzML_Writer":
                software.append(descr_tag)

        soft_list.attrs['count'] = count        

    #Write the new file
    with open(result_file,'w') as file:
        file.write(str(data_need_annotation.prettify()))


def annotate_from_model_imzML(model:str, to_annotate:str):
    """Annotates an imzML file based on an example file - intended for conjunction with write_masked_imzML to preserve metadata"""

    with open(model) as file:
        data = file.read()
    source_imzML = BeautifulSoup(data,'xml')

    with open(to_annotate) as file:
        data = file.read()
    needs_annotation_imzML = BeautifulSoup(data,'xml')

    try:
        instrument_model = source_imzML.instrumentConfigurationList.instrumentConfiguration.attrs['id']
    except:
        instrument_model = "Not found"
    finally:
        needs_annotation_imzML.instrumentConfigurationList.instrumentConfiguration.attrs['id'] = instrument_model

    replace = 'instrumentConfigurationList'
    needs_annotation_imzML.find(replace).replace_with(source_imzML.find(replace))


    new_tag = Tag(builder=needs_annotation_imzML.builder,
                  name="cvParam",
                  attrs={'accession':'MS:1000031',"cvRef":"MS","name":instrument_model})
    needs_annotation_imzML.instrumentConfigurationList.instrumentConfiguration.append(new_tag)

    #Remove empty instrument ref from imzML template
    for paramgroup in needs_annotation_imzML.select("referenceableParamGroupRef"):
        if paramgroup['ref']=="CommonInstrumentParams":
            paramgroup.extract()
    
    for cvParam in needs_annotation_imzML.select("cvParam"):
        if cvParam["accession"]=="MS:1000530":
            del cvParam["value"]
        if cvParam["accession"]=="IMS:1000411":
            cvParam["accession"]="IMS:1000413"
            cvParam["name"]="flyback"
    
    filter_string = source_imzML.find("cvParam", accession="MS:1000512")['value']
    for tag in needs_annotation_imzML.referenceableParamGroupList:
        if "scan1" in str(tag):
            for tag2 in tag:
                if "MS:1000512" in str(tag2):
                    tag2["value"] = filter_string
    
    x_pix_size = source_imzML.find("cvParam", accession="IMS:1000046")['value']
    y_pix_size = source_imzML.find("cvParam", accession="IMS:1000047")['value']
    max_x = source_imzML.find("cvParam", accession="IMS:1000044")['value']
    max_y = source_imzML.find("cvParam", accession="IMS:1000045")['value']

    accessions = ["IMS:1000046", "IMS:1000047", "IMS:1000044", "IMS:1000045"]
    names = ["pixel size x", "pixel size y", "max dimension x", "max dimension y"]
    values = [x_pix_size, y_pix_size, max_x, max_y]

    for i in range(4):
        append_item = f'<cvParam cvRef="IMS" accession="{accessions[i]}" name="{names[i]}" value="{values[i]}"/>\n'
        append_item = BeautifulSoup(append_item,'xml')
        needs_annotation_imzML.scanSettingsList.scanSettings.append(append_item)


    for cvParam in needs_annotation_imzML.select("cvParam"):
        if cvParam["accession"] in accessions:
            cvParam["unitCvRef"]="UO"
            cvParam["unitAccession"]="UO:0000017"
            cvParam["unitName"]="micrometer"

    for soft_list in needs_annotation_imzML.select("softwareList"):
        count = int(soft_list.attrs['count']) + 1
        new_tag = Tag(builder=needs_annotation_imzML.builder,
                    name = "software",
                    attrs = {'id':'imzML_Writer',"version":__version__})
        descr_tag = Tag(builder=needs_annotation_imzML.builder,
                    name = "cvParam",
                    attrs = {'accession':'MS:1000799',"cvRef":"MS","name":"Custom unreleased software tool","value":f"imzML Writer v{__version__}"})

        soft_list.append(new_tag)
        for software in soft_list.select("software"):
            if software.attrs["id"] == "imzML_Writer":
                software.append(descr_tag)

        soft_list.attrs['count'] = count

    #Write the new file
    with open(to_annotate,'w') as file:
        file.write(str(needs_annotation_imzML.prettify())) 
    
    



def write_masked_imzML(source_file:str,roi_mask:np.array,save_dir:str=None) -> str:
    """Rewrites the specified imzML file as masked by the ROI, useful to truncate out only the tissue for reduced file sizes
    
    :param source_file: path to the source imzML
    :param roi_mask: numpy array matching dimensions of source file, where 0 indicates an excluded pixel and 1 is an included pixel
    :param save_dir: Where to save the resulting imzML, defaults to the same directory as the source file
    """

    if not save_dir:
        save_dir = os.path.dirname(source_file)

    filename = Path(source_file).stem
    new_filename = f"{filename}-masked.imzML"
    new_filepath = os.path.join(save_dir,new_filename)

    with warnings.catch_warnings(action='ignore'):
        with imzmlp.ImzMLParser(filename=source_file,parse_lib='lxml') as img:
            total_pixels = len(img.coordinates)
            with imzmlw.ImzMLWriter(new_filepath, mode='processed') as new_imzml:
                zero_offset = 1
                for idx, (x,y,z) in enumerate(img.coordinates):
                    if x == 0 or y == 0:
                        zero_offset = 0

                    if roi_mask[y-zero_offset,x-zero_offset] == 1:
                        mzs, ints = img.getspectrum(idx)
                        new_imzml.addSpectrum(mzs,ints,(x,y,z))
    
    annotate_from_model_imzML(source_file, new_filepath)
    return new_filepath

    
    

    





    



import tkinter as tk
from tkinter import ttk, filedialog,messagebox
import os
import threading
import sys
import time
from importlib import resources

import imzml_writer.imzML_Scout as scout
from imzml_writer.utils import *
from imzml_writer import __version__

timing_mode = False
PC_compiled = False
_on_startup=True
tries = 0

##Colors and FONTS
TEAL = "#2da7ad"
BEIGE = "#dbc076"
GREEN = "#22d10f"
FONT = ("HELVETICA", 18, 'bold')

def gui(tgt_dir:str=None):
    global _on_startup
    """Main control loop for imzML_Writer GUI. No arguments required, but if a directory is passed imzML writer will launch with that directory opened.
    
    :param tgt_dir: (optional) - initial directory for imzML Writer to open in (str)"""

    ##UI Functions
    def get_path():
        """No arguments, prompts the user via dialog box for the directory containing the data to be processed.
        Will call populate_list() method to show files in the UI listbox"""
        global FILE_TYPE
        directory = filedialog.askdirectory(initialdir=os.getcwd())

        if directory:
            CD_entry.delete(0,tk.END)
            CD_entry.insert(0,directory)

            populate_list(directory)
            FILE_TYPE = get_file_types(directory)

            if FILE_TYPE.lower() != "mzML".lower():
                mzML_process.grid_remove()
                imzML_metadata.grid_remove()
            elif FILE_TYPE.lower() == "mzML".lower():
                full_process.grid_remove()
                imzML_metadata.grid_remove()
            elif FILE_TYPE.lower() == "imzML".lower():
                full_process.grid_remove()
                mzML_process.grid_remove()

    def populate_list(dir:str):
        """takes an argument dir and populates the UI listbox based on its contents
        dir: pathname for active directory as a string"""
        file_list.delete(0,tk.END)
        files = os.listdir(dir)
        human_sort(files)
        # files.sort()
        ticker = 0
        for file in files:
            if not file.startswith(".") and not file.endswith(".ibd"):
                if file.endswith(".imzML"):
                    search_txt = file.split(".imzML")[0] + ".ibd"
                    if search_txt in files:
                        file_list.insert(ticker,file)
                    ticker+=1
                else:                
                    file_list.insert(ticker,file)
                    ticker+=1

    def get_file_types(dir) -> str:
        """dir: pathname for active directory
        returns file_type as a str
        [taken as first non-hidden (i.e. doesn't start with ".") file in the directory]"""
        files = os.listdir(dir) 
        for file in files:
            split_file = file.split(".")
            file_type = split_file[-1]

        file_type_label = tk.Label(text=f"File type: .{file_type}",bg=TEAL,font=FONT)
        file_type_label.grid(row=1,column=3,columnspan=3)
        return file_type


    def full_convert():
        """Initiates file conversion from vendor format in the current directory"""
        if timing_mode:
            global tic
            tic = time.time()
        #RAW to mzML conversion, then call mzML to imzML function
        file_type = get_file_type(CD_entry.get())
        msconvert_call = threading.Thread(target=RAW_to_mzML,kwargs={"path":CD_entry.get(),"write_mode":write_option_var.get(), "combine_ion_mobility": combine_ion_mobility.get()})
        msconvert_call.start()

        RAW_progress.config(mode="indeterminate")
        RAW_progress.start()
        follow_raw_progress(file_type,msconvert_call)

    def follow_raw_progress(raw_filetype:str,convert_thread:threading.Thread):
        """Monitors progress of raw file conversion to mzML by comparing the number of raw vendor files to mzML files in the working directory
        Input:
        raw_filetype: string specifying the file extension of the raw files"""
        global tic

        still_active = convert_thread.is_alive()

        #Retrieve list of files in working directory
        files = os.listdir(CD_entry.get())
        num_raw_files = 0
        num_mzML_files = 0
        mzML_files = []

        ##Iterate through each file, counting each type
        for file in files:
            if file.startswith(".")==False:
                if f".{raw_filetype}" in file:
                    num_raw_files+=1
                elif "mzML" in file:
                    num_mzML_files+=1
                    mzML_files.append(file)
        
        #Calculate progress based on number of each
        progress = int(num_mzML_files * 100 / num_raw_files)
        
                
        #Update progress bar to show how many mzML files are finished compared to total
        if progress > 0:
            RAW_progress.stop()
            RAW_progress.config(mode="determinate",value=progress)

        #If not finished, start this function over again after waiting 3 seconds for more progress to be made
        if progress < 100 or still_active:
            window.after(3000,lambda:follow_raw_progress(raw_filetype, convert_thread))
        #If finished, move on to the next stage in the process
        elif progress >= 100 and not still_active:

            
            if timing_mode:
                global tic
                toc = time.time()
                print(f"RAW to mzML: {round(toc - tic,1)}s")
            #Clean up file structure by placing mzML and raw files in separate folders
            clean_raw_files(path=CD_entry.get(),file_type=raw_filetype)
            #Make it obvious the process is complete by changing the label to green
            RAW_label.config(fg=GREEN)

            ##Change the directory to the new mzML folder
            new_path = os.path.join(CD_entry.get(),"Output mzML Files")
            CD_entry.delete(0,tk.END)
            CD_entry.insert(0,new_path)
            populate_list(CD_entry.get())

            ##Initiate the next step in the pipeline
            window.after(500,mzML_to_imzML())

        
    def mzML_to_imzML():
        """Run main conversion script from mzML to imzML, stop at annotation stage"""
        cur_path = CD_entry.get()

        ##Retrieve settings
        duplicate_spectra = duplicate_bool.get()
        zero_indexed = index_bool.get()
        search_tolerance = search_tol.get()
        try:
            search_tolerance = float(search_tolerance)
        except:
            search_tolerance = 20
            messagebox.showwarning(title="ERROR",message="Invalid search tolerance specified - proceeding with 20 ppm")

        ##Start the progress bar whirling to indicate to user that things are working
        write_imzML_progress.config(mode="indeterminate")
        write_imzML_progress.start()
        if os.path.basename(cur_path) != "Output mzML Files":
            clean_raw_files(cur_path,"     ")
            cur_path = os.path.join(cur_path,"Output mzML Files")
        

        ##Start thread to convert the process
        thread = threading.Thread(
            target=lambda:mzML_to_imzML_convert(
                PATH=cur_path,
                progress_target=write_imzML_progress,
                LOCK_MASS=lock_mass_entry.get(),
                TOLERANCE=search_tolerance,
                no_duplicating=duplicate_spectra,
                zero_indexed=zero_indexed))
        thread.daemon=True
        thread.start()

        ##Start monitoring process to see if imzML files have been successfully written
        check_imzML_completion(thread)
        
    def check_imzML_completion(thread):
        """monitors imzML conversion process by checking if the thread is still alive"""
        if thread.is_alive():
            window.after(2000,check_imzML_completion,thread) #If thread is still going, check back again in 2 seconds
        else: #Otherwise, move on to the next step
            if timing_mode:
                global tic
                toc = time.time()
                print(f"mzML to imzML: {round(toc - tic,1)}s")
            #Update progress bar label to green to make it obvious things have completed
            write_imzML_Label.config(fg=GREEN)

            ##Update folder to directory where the intermediate imzML files are saved (directory w/ the python code)
            full_path = CD_entry.get()
            new_path = os.path.dirname(full_path)
            CD_entry.delete(0,tk.END)
            CD_entry.insert(0,new_path)
            populate_list(os.getcwd())

            ##Initiate the metadata writing process
            write_metadata(path_in="indirect")

    def write_metadata(path_in:str="direct"):
        """Initiates metadata writing on the intermediate mzML files"""
        global path_to_models

        #Start progress bar whirring to indicate process has started to user
        Annotate_progress.config(mode="indeterminate")
        Annotate_progress.start()
        #If conversion was called directly, prompt user for source mzml files to retrieve metadata from
        if path_in == "direct":
            path_to_models = filedialog.askdirectory(initialdir=os.getcwd())
        else:
            path_to_models = os.path.join(CD_entry.get(),"Output mzML Files")

        
        #Start the annotation in a new thread
        thread = threading.Thread(
            target=lambda:imzML_metadata_process(
                model_files=path_to_models,
                x_speed=int(speed_entry.get()),
                y_step=int(Y_step_entry.get()),
                tgt_progress=Annotate_progress,
                path=CD_entry.get()))
        thread.daemon=True
        thread.start()

        ##Monitor the annotation
        check_metadata_completion(thread)

    def check_metadata_completion(thread):
        """Follows metadata writing process, moving on if thread has terminated or checking again if it hasn't"""
        global path_to_models
        ##If thread is still going, run this function again after waiting 2 seconds
        if thread.is_alive():
            window.after(2000,check_metadata_completion,thread)
        else: ##Otherwise, move on
            if timing_mode:
                global tic
                toc = time.time()
                print(f"imzML metadata: {round(toc - tic,1)}s")
            ##Update label to green, update file list to indicate process completion to user
            Annotate_recalibrate_label.config(fg=GREEN)
            model_file_list = os.listdir(path_to_models)
            model_file_list.sort()

            str_array = [letter for letter in model_file_list[0]]
            OUTPUT_NAME = "".join(str_array)
            while OUTPUT_NAME not in model_file_list[-1]:
                str_array.pop(-1)
                OUTPUT_NAME = "".join(str_array)

            new_path = f"{CD_entry.get()}/{OUTPUT_NAME}"
            CD_entry.delete(0,tk.END)
            CD_entry.insert(0,new_path)
            populate_list(CD_entry.get())
                
            

    def launch_scout():
        """Launches imzML_Scout.py when the user selects an imzML file on the indicated file"""
        if PC_compiled:
            messagebox.showwarning(title="imzML Scout Unavailable...",message="On PC, imzML Scout is only available using the python distributable from pypi. See the Github page for installation instructions.")
        else:
            tgt_file = file_list.selection_get()
            if tgt_file.split(".")[-1]=="ibd":
                file_start = tgt_file.split("ibd")[0]
                tgt_file = file_start+"imzML"
            path = CD_entry.get()
            file_path = f"{path}/{tgt_file}"
            scout.main(tgt_file=file_path)
        

    def resource_path(relative_path):
        """Future placeholder for making standalone application work"""
        try:
            base_path = sys._MEIPASS
        except Exception:
            base_path = os.path.abspath(".")

        return os.path.join(base_path, relative_path)

    
    def launch_advanced():
        def update_search_tol(*args):
            try:
                float(lock_mass_search_tol_entry.get())
                search_tol.set(lock_mass_search_tol_entry.get())
            except:
                messagebox.showwarning(title="ERROR",message="Search tolerance must be specified as number")
        
        def update_index_bool(*args):
            if not index_bool.get():
                index_bool.set(True)
            else:
                index_bool.set(False)
        
        def update_duplicate_bool(*args):
            if not duplicate_bool.get():
                duplicate_bool.set(True)
            else:
                duplicate_bool.set(False)
        
        def update_mobility_bool(*args):
            if not combine_ion_mobility.get():
                combine_ion_mobility.set(True)
            else:
                combine_ion_mobility.set(False)

        advanced_window = tk.Tk()
        advanced_window.title("Advanced Options...")
        advanced_window.config(padx=5,pady=5,bg=TEAL)

        #0 vs 1 indexed (checkbox, 1 default, 0 optional)
        index_check = tk.Checkbutton(advanced_window,text="0-Indexed?",bg=TEAL,font=FONT,variable=index_bool,command=update_index_bool)
        index_check.grid(row=1,column=1,columnspan=2)
        if index_bool.get():
            index_check.select()

        #Duplicate pixels? (checkbox, duplicate default)
        duplicate_check = tk.Checkbutton(advanced_window,text="no duplicated spectra?",bg=TEAL,font=FONT,command=update_duplicate_bool)
        duplicate_check.grid(row=2,column=1,columnspan=2)
        if duplicate_bool.get():
            duplicate_check.select()
        
        #Combine ion mobility spectra
        combine_ion_mob_entry = tk.Checkbutton(advanced_window,text="Combine ion mobility spectra?", bg=TEAL, font=FONT,command=update_mobility_bool)
        combine_ion_mob_entry.grid(row=3,column=1,columnspan=2)
        if combine_ion_mobility.get():
            combine_ion_mob_entry.select()


        #Lock mass search tolerance (entry, 20 ppm default)
        lock_mass_search_tol_label = tk.Label(advanced_window,text="Lock mass search tolerance (ppm):",bg=TEAL,font=FONT)
        lock_mass_search_tol_entry = tk.Entry(advanced_window,text="Enter tolerance here",highlightbackground=TEAL,background=BEIGE,fg="black",justify='center')
        lock_mass_search_tol_entry.insert(0,search_tol.get())
        lock_mass_search_tol_entry.bind("<Return>",update_search_tol)
        lock_mass_search_tol_entry.bind("<FocusOut>",update_search_tol)

        lock_mass_search_tol_entry.grid(row=4,column=2)
        lock_mass_search_tol_label.grid(row=4,column=1)


            


    ##Build tkinter window
    window = tk.Tk()
    window.title(f"imzML Writer v{__version__}")
    window.config(padx=5,pady=5,bg=TEAL)
    style = ttk.Style()
    style.theme_use('clam')



    ##Logo
    try:
        canvas = tk.Canvas(width = 313,height=205,bg=TEAL,highlightthickness=0)
        img = tk.PhotoImage(file=resource_path("Images/Logo-01.png"))
        window.iconbitmap(resource_path("Images/imzML_Writer.ico"))
        canvas.create_image(313/2, 205/2,image=img)
        canvas.grid(column=0,row=0,columnspan=2)
    except:
        try:
            canvas = tk.Canvas(width = 313,height=205,bg=TEAL,highlightthickness=0)
            with resources.path('imzml_writer.Images','Logo-01.png') as path:
                img=tk.PhotoImage(file=resource_path(path))
                canvas.create_image(313/2, 205/2,image=img)
                canvas.grid(column=0,row=0,columnspan=2)
            with resources.path('imzml_writer.Images','imzML_Writer.ico') as path:
                window.iconbitmap(resource_path(path))
        except:
            pass
    
    ##Initialize defaults for advanced options
    search_tol = tk.StringVar(window)
    search_tol.set("20")
    duplicate_bool = tk.BooleanVar(window)
    duplicate_bool.set(False)
    index_bool = tk.BooleanVar(window)
    index_bool.set(False)
    combine_ion_mobility = tk.BooleanVar(window)
    combine_ion_mobility.set(False)

    ##Scan-speed entry
    speed_label=tk.Label(text="x scan speed (µm/s):",bg=TEAL,font=FONT)
    speed_entry = tk.Entry(text="Enter speed here",highlightbackground=TEAL,background=BEIGE,fg="black",justify='center')
    speed_entry.insert(0,"40")

    speed_label.grid(row=2,column=0)
    speed_entry.grid(row=2,column=1)

    ##Y-step entry
    Y_step_label=tk.Label(text="y step (µm):",bg=TEAL,font=FONT)
    Y_step_entry=tk.Entry(highlightbackground=TEAL,background=BEIGE,fg="black",justify='center')
    Y_step_entry.insert(0,"150")

    Y_step_label.grid(row=3,column=0)
    Y_step_entry.grid(row=3,column=1)

    ##Lock mass entry
    lock_mass_label=tk.Label(text="Lock Mass:",bg=TEAL,font=FONT)
    lock_mass_entry=tk.Entry(highlightbackground=TEAL,background=BEIGE,fg="black",justify='center')
    lock_mass_entry.insert(0,"0")

    lock_mass_label.grid(row=4,column=0)
    lock_mass_entry.grid(row=4,column=1)

    ##Choose Directory Button
    CD_button = tk.Button(text="Select Folder",bg=TEAL,highlightbackground=TEAL,command=get_path)
    CD_button.grid(row=1,column=0)

    CD_entry = tk.Entry(text="Enter Directory Here",highlightbackground=TEAL,background=BEIGE,fg="black",justify='center')
    CD_entry.grid(row=1,column=1)

    ##RAW conversion progress bar
    RAW_label = tk.Label(text="RAW --> mzML:",bg=TEAL,font=FONT)
    RAW_label.grid(row = 5,column=0)

    RAW_progress = ttk.Progressbar(length=525,style="danger.Striped.Horizontal.TProgressbar")
    RAW_progress.grid(row=5,column=1,columnspan=5)


    ##Write imzML progress bar
    write_imzML_Label=tk.Label(text="Write imzML:",bg=TEAL,font=FONT)
    write_imzML_Label.grid(row=6,column=0)

    write_imzML_progress=ttk.Progressbar(length=525,style="info.Striped.Horizontal.TProgressbar")
    write_imzML_progress.grid(row=6,column=1,columnspan=5)


    ##Annotate / m/z recalibration progress bar:
    Annotate_recalibrate_label = tk.Label(text="Metadata:",bg=TEAL,font=FONT)
    Annotate_recalibrate_label.grid(row=7,column=0)

    Annotate_progress=ttk.Progressbar(length=525,style="success.Striped.Horizontal.TProgressbar")
    Annotate_progress.grid(row=7,column=1,columnspan=5)



    #Listbox for files in target folder
    file_list = tk.Listbox(window,bg=BEIGE,fg="black",height=10,highlightcolor=TEAL,width=35,justify='left')
    file_list.grid(row=0,column=4,rowspan=2,columnspan=3)

    ##Processing buttons
    full_process = tk.Button(text="Full Conversion",bg=TEAL,highlightbackground=TEAL,command=full_convert)
    full_process.grid(row=2,column=4)

    mzML_process = tk.Button(text="mzML to imzML",bg=TEAL,highlightbackground=TEAL,command=mzML_to_imzML)
    mzML_process.grid(row=2,column=5)

    imzML_metadata = tk.Button(text="Write imzML metadata",bg=TEAL,highlightbackground=TEAL,command=write_metadata)
    imzML_metadata.grid(row=3,column=4)

    #Advanced option
    adv_options = tk.Button(text="Advanced Options...",bg=TEAL,highlightbackground=TEAL,command=launch_advanced)
    adv_options.grid(row=3,column=5)
    

    ##Visualize .imzML
    visualize = tk.Button(text="View imzML",bg=TEAL,highlightbackground=TEAL,command=launch_scout)
    visualize.grid(row=4,column=5,columnspan=2)

    ##Centroid or Profile?
    data_writing_options = ["Centroid", "Profile"]
    write_option_var = tk.StringVar(window)
    write_option_var.set(data_writing_options[0])
    write_options_dropdown=tk.OptionMenu(window,write_option_var,*data_writing_options)
    write_options_dropdown.grid(row=4,column=4)

    if _on_startup:
        _on_startup = False
        if tgt_dir != None:
            CD_entry.delete(0,tk.END)
            CD_entry.insert(0,tgt_dir)

            populate_list(tgt_dir)
            FILE_TYPE = get_file_types(tgt_dir)

            if FILE_TYPE.lower() != "mzML".lower():
                mzML_process.grid_remove()
                imzML_metadata.grid_remove()
            elif FILE_TYPE.lower() == "mzML".lower():
                full_process.grid_remove()
                imzML_metadata.grid_remove()
            elif FILE_TYPE.lower() == "imzML".lower():
                full_process.grid_remove()
                mzML_process.grid_remove()

    window.mainloop()

if __name__=="__main__":
    try:
        gui(sys.argv[1])
    except:
        gui()


import pyimzml.ImzMLParser as imzmlp
import matplotlib.pyplot as plt
import tkinter as tk
from tkinter import ttk, filedialog, messagebox
from RangeSlider.RangeSlider import RangeSliderV
import os
import sys
from matplotlib.figure import Figure 
from matplotlib.backends.backend_tkagg import (FigureCanvasTkAgg, NavigationToolbar2Tk)
from matplotlib.backend_tools import ToolBase, ToolToggleBase 
import warnings
import numpy as np
import pandas as pd
import json

from imzml_writer import scout_utils
from imzml_writer.analyte_list_cleanup import *
from imzml_writer import __version__



def main(tgt_file:str = "",initial_mz:float=104.1070):
    """Main control loop for imzML Scout GUI. Callable either with no arguments (find file via GUI) or by passing the file path to
    the target imzML.
    
    :param tgt_file: Path to imzML file for visualization."""
    ##Colors and FONTS
    TEAL = "#2da7ad"
    BEIGE = "#dbc076"
    FONT = ("HELVETICA", 18, 'bold')


    def getionimage_wrapper(img_file:imzmlp.ImzMLParser,tgt_mz:float,window:float):
        image = imzmlp.getionimage(p=img_file, mz_value=tgt_mz, tol=window)
        if tgt_mz >= low_range and tgt_mz <= high_range:
            return image
        else:
            dims = image.shape
            image = np.zeros(shape=dims)
            return image

    def browse_for_file():
        """Launch dialog box for user to find and select a target imzML file"""
        path_to_file = filedialog.askopenfilename(initialdir=os.getcwd(),filetypes=[("imzML Files","*.imzML")])
        file_entry.delete(0,tk.END)
        file_entry.insert(0,path_to_file)

    def get_aspect_ratio(img_file):
        """Tries to extract aspect ratio information (relative pixel dimensions) from the imzML. Uses 1x1 pixels if it fails"""
        try:
            metadata = img_file.metadata.pretty()
            x_pix_size = metadata["scan_settings"]["scanSettings1"]["pixel size (x)"]
            y_pix_size = metadata["scan_settings"]["scanSettings1"]["pixel size y"]

            max_x_dimension = metadata["scan_settings"]["scanSettings1"]["max dimension x"]

            return y_pix_size/x_pix_size, x_pix_size, y_pix_size, max_x_dimension
        except:
            return 1, 1, 1, 1
    
    def get_scan_range(img_file:imzmlp.ImzMLParser):
        global low_range, high_range
        first_scan = img_file.getspectrum(0)
        low_range = np.min(first_scan[0])
        high_range = np.max(first_scan[0])
        


    def plot_ion_image(*_):
        """Deals with data update for plotting a new ion image, retrieving targets from the user input fields"""
        global raw_ion_image, aspect_ratio,x_pix,y_pix, imzML_object, max_x_dimension

        #Get the target data for the iamge
        target_mz = float(mz_entry.get())
        tolerance = float(tolerance_entry.get())
        filename = file_entry.get()
        mz_window=target_mz*tolerance/1e6

        #Open the imzML datafile if needed
        with warnings.catch_warnings(action="ignore"):
            # coordinate_map = []
            if "imzML_object" in globals():
                if imzML_object.filename != filename:
                    imzML_object = imzmlp.ImzMLParser(filename=filename,parse_lib='lxml')
            else:
                imzML_object = imzmlp.ImzMLParser(filename=filename,parse_lib='lxml')
        
        
        get_scan_range(imzML_object)

        #Check if TIC image was requested, view entire spectrum if so. Otherwise generate datagrid of requested m/z and tolerance combo
        if view_tic_option.get():
            ion_image = getionimage_wrapper(imzML_object,mz_value=200,tol=9999)
        else:
            ion_image = getionimage_wrapper(imzML_object,target_mz,mz_window)
        
        #Get aspect ratio
        [aspect_ratio, x_pix, y_pix, max_x_dimension] = get_aspect_ratio(imzML_object)

        #Normalize ion image (or don't) as specified in GUI
        norm_method = normalization_method.get()
        if norm_method == "custom":
            norm_mz = float(normalize_custom_entry.get())
            norm_window = norm_mz * tolerance / 1e6
            norm_grid = getionimage_wrapper(imzML_object,mz_value=norm_mz,tol=norm_window)
            ion_image = np.divide(ion_image,norm_grid,out=np.zeros_like(ion_image),where=norm_grid!=0)
        elif norm_method == "TIC":
            norm_grid = getionimage_wrapper(imzML_object,mz_value=200,tol=9999)
            ion_image = np.divide(ion_image,norm_grid,out=np.zeros_like(ion_image),where=norm_grid!=0)
        
        #Initiate new raw data variable so we can freely manipulate the ion image as needed
        raw_ion_image = ion_image

        fig = update_ion_image()
        return fig

    def update_ion_image(*_):
        """Updates the actual ion image in the GUI, with various checks for user input settings and handling to remove old images"""
        global raw_ion_image,aspect_ratio,x_pix,y_pix,canvas_ionimage,title_label,fig,ion_image, plot1, last_selected_patch, last_selected_pixel, red_highlight_patch, color_NL

        ##Retrieve contrast cutoffs for low-end and top-end
        low_thres = v_bottom.get()
        up_thres=v_top.get()
        target_mz = float(mz_entry.get())
        tolerance = float(tolerance_entry.get())

        ion_image = raw_ion_image
        if apply_ROI_mask.get():
            ion_image = ion_image * scout_utils.roi_mask
        
        #Check is smoothing is to be applied, if so do it
        if smooth_state.get():
            ion_image = scout_utils.smooth_image(ion_image, aspect_ratio)

        ##Apply the cutoffs, setting those below to 0 and those above to high_cutoff value
        low_cutoff = np.percentile(ion_image,low_thres*100)
        up_cutoff = np.percentile(ion_image,up_thres*100)

        ion_image = np.where(ion_image > up_cutoff,up_cutoff,ion_image)
        ion_image = np.where(ion_image < low_cutoff,0,ion_image)

        #If custom normalization set, apply it to the colormap
        if NL_state.get():
            color_NL = norm_value.get()
        else:
            color_NL= up_cutoff

        


        ##Initiate and raw ion image
        fig = Figure(dpi=100,facecolor=TEAL,layout='tight')
        
        plot1 = fig.add_subplot()
        plot1.imshow(ion_image,aspect=aspect_ratio,interpolation="none",vmin=0,vmax=color_NL,cmap=cmap_selected.get())
        plot1.axis('off')

        #Remove old image objects before inserting a new one
        if not first_img.get():
            canvas_ionimage.get_tk_widget().destroy()
            title_label.destroy()
        else:
            last_selected_pixel = None
            last_selected_patch = None
            red_highlight_patch = None

        #Add the ion image to the tkinter window
        canvas_ionimage = FigureCanvasTkAgg(fig,master=window_scout)
        
        #Draw the selected patch if available on update of the ion image
        if last_selected_patch != None:
            draw_last_selected_patch()
        canvas_ionimage.draw()
        toolbar = NavigationToolbar2Tk(canvas_ionimage, pack_toolbar=False)
        toolbar.update()
        canvas_ionimage.get_tk_widget().grid(row=5,column=0,columnspan=3)

        ##Draw a label for the ion image with target m/z, tolerance, and pixel dimensions
        title_string=[]
        title_string = f"{int(round(x_pix,0))} µm x {int(round(y_pix,1))} µm pixels; m/z {target_mz} @ {tolerance} ppm"
        title_label = tk.Label(window_scout,text=title_string,bg=TEAL,font=FONT)
        title_label.grid(row=6,column=0,columnspan=4)

        #Initiate callbacks for when users mouse over the ion image for viewing/selecting highlighted pixel
        fig.canvas.mpl_connect("motion_notify_event",image_move)
        fig.canvas.callbacks.connect('button_press_event',report_coordinates)
        first_img.set(False)
        
        return fig

    def draw_last_selected_patch():
        """Draws a red patch wherever the last selected pixel was"""
        global last_selected_patch, last_selected_pixel
        lx, ly = last_selected_pixel

        # Prepare the green overlay for the selected pixel
        selected_overlay = np.zeros((ion_image.shape[0], ion_image.shape[1], 4))  # New overlay
        
        num_lines, pixels_per_line = ion_image.shape
        selected_min_y = int(max(0, ly))
        selected_max_y = int(min(num_lines, ly + 1))
        selected_min_x = int(max(0, lx - pixels_per_line*0.008))
        selected_max_x = int(min(pixels_per_line, lx + 1))


        # Set the green color with some opacity for the selection overlay
        selected_overlay[selected_min_y:selected_max_y, selected_min_x:selected_max_x] = [1, 0, 0, 0.8]  # Green with 80% opacity

        last_selected_patch = plot1.imshow(selected_overlay, aspect=aspect_ratio, interpolation="none")
        
    def image_move(event):
        """Handle mouse movement over the ion image and raws a green patch over the currently selectable region."""
        global ion_image, plot1, last_selected_patch, red_highlight_patch,canvas_ionimage, color_NL  # Track selected and red patches

        if event.xdata is not None and event.ydata is not None:
            # Get the pixel index
            x_index = int(event.xdata)
            y_index = int(event.ydata)

            # Create a copy of the original image to modify for hover
            highlight_image = np.zeros((ion_image.shape[0], ion_image.shape[1], 4))  # RGBA image for highlight

            # Define bounds for the highlight area
            pixels_per_line = ion_image.shape

            min_y = int(max(0, y_index))
            max_y = int(min(ion_image.shape[0], y_index+1))
            min_x = int(max(0, x_index - pixels_per_line[1]*0.008))
            max_x = int(min(ion_image.shape[1], x_index + 1))

            # Set the green color with varying opacity for the hover overlay
            highlight_image[min_y:max_y, min_x:max_x] = [0, 1, 0, 0.8]  # Green with 80% opacity        
            
            # Draw the ion image first
            plot1.imshow(ion_image, aspect=aspect_ratio, interpolation="none", vmin=np.min(ion_image), vmax=color_NL, cmap=cmap_selected.get())

            # Overlay the red highlight
            if red_highlight_patch is not None:
                red_highlight_patch.remove()  # Remove previous red highlight if it exists
                
            red_highlight_patch = plot1.imshow(highlight_image, aspect=aspect_ratio, interpolation="none")  # Store new red highlight

            # If we have a last selected pixel, keep it drawn
            if last_selected_pixel:
                draw_last_selected_patch()

            # Refresh the canvas to show updates
            canvas_ionimage.draw()

    def export_csv():
        """Exports the current image data as a csv file, prompts the user as to where they'd like to save it."""
        global raw_ion_image
        dataframe = pd.DataFrame(raw_ion_image)
        path = os.path.dirname(file_entry.get()) 
        file_name = filedialog.asksaveasfilename(initialdir=path,filetypes=[("CSV file",".csv")])
        dataframe.to_csv(path_or_buf=file_name,header=False,index=False)

    def bulk_export_csv():
        """Exports a batch of images as csv files, prompts the user for a spreadsheet of target m/z and names"""
        global raw_ion_image

        ##Prompt user for a spreadsheet of target m/z and names to give them
        target_list_file = filedialog.askopenfilename(initialdir=os.getcwd(),filetypes=[("Excel Spreadsheet",".xlsx"),("CSV File",".csv")])
        target_list = pd.read_excel(target_list_file)
        target_list=cleanup_table(target_list,target_list_file)

        ##Generate ion image for each entry, write the raw data to a CSV file
        for iter,row in target_list.iterrows():
            mz_entry.delete(0,tk.END)
            mz_entry.insert(0,row.values[1])
            plot_ion_image()

            folder_name = os.path.join(os.path.dirname(file_entry.get()),"ion_images")
            img_name_base = f"{row.values[0]}-{str(row.values[1]).split(".")[0]}"
            if iter == 0:
                if os.path.exists(folder_name):
                    messagebox.showwarning(title="Folder already exists!",message="You already have an ion image folder here, please rename, move, or delete it")
                    break
                os.mkdir(folder_name)
                # file = filedialog.asksaveasfilename(initialdir=folder_name,filetypes=[("CSV", ".csv")],initialfile=img_name_base)
                used_extension="csv"
                # used_extension = file.split(".")[-1]
                # if file.endswith("csv") or file.endswith("CSV"):
                #     file = file + ".csv"

            file = os.path.join(folder_name,f"{img_name_base}.{used_extension}")
            
            dataframe = pd.DataFrame(raw_ion_image)
            dataframe.to_csv(path_or_buf=file,header=False,index=False)

        ##Optionally, write the TIC image as well
        if include_TIC_var.get():
            view_tic_check.invoke()
            file = os.path.join(folder_name,f"TIC_Image.{used_extension}")
            dataframe = pd.DataFrame(raw_ion_image)
            dataframe.to_csv(path_or_buf=file,header=False,index=False)

    def find_scan_idx(event):
        """Based on where the user clicks, find the corresponding scan index in the imzML file"""
        factor = 3
        if smooth_state.get():
            x_coord = int(event.xdata / factor) + 1
            y_coord = int(event.ydata / factor) + 1
        else:
            x_coord = int(event.xdata) + 1
            y_coord = int(event.ydata) + 1
        
        search_coords = (x_coord, y_coord, 1)
        if search_coords in imzML_object.coordinates:
            return imzML_object.coordinates.index(search_coords)
        else:
            return None


    def hide_patches():
        global red_highlight_patch, last_selected_patch
        if red_highlight_patch != None:
            red_highlight_patch.set_visible(False)

        
        if last_selected_patch !=None:
            last_selected_patch.set_visible(False)

    def show_patches():
        global red_highlight_patch, last_selected_patch
        if red_highlight_patch != None:
            red_highlight_patch.set_visible(True)

        
        if last_selected_patch !=None:
            last_selected_patch.set_visible(True)


    def export_image(fig):
        """Export the currently viewed image as an image file (tif, png, jpg), prompt the user for where to put it"""

        hide_patches()
        
        #Prompt the user for where to save it
        file = filedialog.asksaveasfilename(initialdir=os.getcwd(),filetypes=[("TIF", ".tif"),("PNG",".png"),("JPG", ".jpg")])
        quality = qual_var.get()
        try:
            quality = float(quality)
        except:
            messagebox.showwarning(title="Invalid dpi setting...",message="dpi should be specified as a number - proceeding with default (dpi = 100)")
            quality = 100
        
        if file:
            #Save the file
            file_format = file.split(".")[-1]
            if file_format not in ["tif", "jpg", "png","tiff"]:
                file_format = "tif"
                file = f"{file}.tif"
            
            fig.savefig(fname=file,
                        transparent=True,
                        dpi=quality,
                        format=file_format,
                        bbox_inches="tight",
                        pad_inches=0)
        
        show_patches()
            


    def check_normalization():
        """Handles contextual display on whether a custom normalization m/z is being applied"""
        if normalization_method.get() == "custom": ##If custom NL, add an entry for that m/z and listeners to update the ion image when it changes
            normalize_custom_entry.grid(row=2,column=6)
            normalize_custom_entry.bind("<Return>",plot_ion_image)
            normalize_custom_entry.bind("<FocusOut>",plot_ion_image)
        else:
            try:
                normalize_custom_entry.grid_remove() #Remove the box for space if the box is unchecked
            except:
                pass
        
        try:
            plot_ion_image()
        except:
            pass
        
    def report_coordinates(event):
        """Reports the current pixel x/y and updates the mass spectrum from the scan_idx"""
        global scan_idx, last_selected_pixel
        if event.xdata != None:
            scan_idx = find_scan_idx(event)
            if scan_idx != None:
                last_selected_pixel = (event.xdata, event.ydata)
                update_plot_for_selected_pixel()
                plot_mass_spectrum(scan_idx)
        
    def update_plot_for_selected_pixel():
        """Update the plot with the new selected pixel and draw the previous selection in red."""
        global ion_image, plot1, canvas_ionimage, last_selected_pixel, last_selected_patch

        # Draw the ion image
        plot1.imshow(ion_image, aspect=aspect_ratio, interpolation="none", vmin=np.min(ion_image), vmax=color_NL, cmap=cmap_selected.get())
        
        # Remove old highlights
        if last_selected_patch is not None:
            last_selected_patch.remove()

        # Add red highlight, if a pixel is selected
        if last_selected_pixel:
            draw_last_selected_patch()

        # Refresh the canvas
        canvas_ionimage.draw()


    def plot_mass_spectrum(scan_idx):
        """Plot a mass spectrum and add listeners to watch for mouseovers/clicks to update the ion image to the nearest detectable m/z"""
        global imzML_object, mz, intensities, MS_vline, plot2, canvas_mass_spectrum

        ##Retrieve the data for the target spectrum
        [mz, intensities] = imzML_object.getspectrum(scan_idx)

        ##Plot the actual mass spectrum using vlines
        fig_spectrum = Figure(figsize=(4,4),dpi=100,facecolor=TEAL)
        plot2 = fig_spectrum.add_subplot()
        plot2.vlines(x=mz,ymin=0,ymax=intensities)
        plot2.set_ylim(0,plot2.get_ylim()[1])
        if not first_MS.get():
            #Set the x lims and y lims, as input by the user
            plot2.set_xlim(float(start_var.get()),float(end_var.get()))
            in_bounds = [idx for idx, value in enumerate(mz) if float(start_var.get()) < value < float(end_var.get())]
            new_ylim = np.max(intensities[in_bounds]) *1.3
            plot2.set_ylim(0,new_ylim)
            
        plot2.set_xlabel("m/z")
        plot2.set_ylabel("Intensity")

        try:
            ##Remove old mass spectra so they aren't layered on top of each other
            canvas_mass_spectrum.destroy()
        except:
            pass
        
        ##Add mass spectrum to the GUI window
        canvas_mass_spectrum = FigureCanvasTkAgg(fig_spectrum,master=window_scout)
        canvas_mass_spectrum.get_tk_widget().grid(row=5,column=6,columnspan=4)


        if first_MS.get():
            ##Handling for start/stop xlims entries, listeners to update when these are changed, etc.
            MS_vline = None
            start_mz = tk.Label(window_scout,text="Start m/z",bg=TEAL,font=FONT)
            end_mz = tk.Label(window_scout, text = "End m/z",bg=TEAL,font=FONT)
            start_mz.grid(row=6,column=6)
            end_mz.grid(row=6,column=8)
            start_var.set(f"{plot2.get_xlim()[0]:.1f}")
            start_mz_entry = tk.Entry(window_scout,textvariable=start_var,highlightbackground=TEAL,background=BEIGE,fg="black",justify='center',width=10)
            end_var.set(f"{plot2.get_xlim()[1]:.1f}")
            end_mz_entry = tk.Entry(window_scout,textvariable=end_var,highlightbackground=TEAL,background=BEIGE,fg='black',justify='center',width=10)
            start_mz_entry.grid(row=6,column=7)
            end_mz_entry.grid(row=6,column=9)
            start_mz_entry.bind("<Return>",lambda event:plot_mass_spectrum(scan_idx=scan_idx))
            start_mz_entry.bind("<FocusOut>",lambda event:plot_mass_spectrum(scan_idx=scan_idx))
            end_mz_entry.bind("<Return>",lambda event:plot_mass_spectrum(scan_idx=scan_idx))
            end_mz_entry.bind("<FocusOut>",lambda event:plot_mass_spectrum(scan_idx=scan_idx))

        ##Show previously selected m/z as needed
        current_mz = mz_entry.get()
        current_mz = float(current_mz)
        xlim = plot2.get_xlim()
        percentage = (current_mz - xlim[0]) / (xlim[1] - xlim[0])
        if percentage > 0.75:
            ha = 'right'  # align label to the right if near the end
            label_x_pos = current_mz - 0.5  # adjust label position to the left
        else:
            ha = 'left'   # align label to the left otherwise
            label_x_pos = current_mz + 0.5
        
        y_position = 4/5 * plot2.get_ylim()[1]
        plot2.axvline(x=current_mz,color='red',linestyle='--')
        plot2.text(label_x_pos, y_position, f"{current_mz:.4f}", color='red', fontsize=10, ha=ha, va='center')
        canvas_mass_spectrum.draw()

        ##Add listeners for mouseover and click on the mass spectrum
        fig_spectrum.canvas.mpl_connect("motion_notify_event",on_MS_move)
        fig_spectrum.canvas.callbacks.connect("button_press_event",change_target_mz)
        first_MS.set(False)
    
    def on_MS_move(event):
        """Listener for mouseover events on the mass spectrum, finds the nearest m/z when you do so and displays it on the window"""
        new_mz = event.xdata
        if new_mz is not None:
            new_target = find_target_mz(new_mz)

            global MS_vline, plot2, canvas_mass_spectrum, MS_label
            if MS_vline is not None:
                MS_vline.remove()

            if 'MS_label' in globals() and MS_label is not None:
                MS_label.remove()

            xlim = plot2.get_xlim()
            percentage = (new_target - xlim[0]) / (xlim[1] - xlim[0]) #percentage along x-axis
            y_position = 2 / 3 * plot2.get_ylim()[1]

            # Contextual align label left/right of the bar depending on how close to the end of viewing window we are
            if percentage > 0.75:
                ha = 'right'  # align label to the right if near the end
                label_x_pos = new_target - 0.5  # adjust label position to the left
            else:
                ha = 'left'   # align label to the left otherwise
                label_x_pos = new_target + 0.5  # adjust label position to the right

            MS_vline = plot2.axvline(x=new_target, color='black', linestyle='--')
            
            # Create the label with adjusted properties
            MS_label = plot2.text(label_x_pos, y_position, f"{new_target:.4f}",
                                color='black', fontsize=10, ha=ha, va='center')

            canvas_mass_spectrum.draw()

    def find_target_mz(new_mz):
        """Finds the target m/z value based on surrounding, detectable m/z values."""
        if new_mz is None:
            return None  # Handle case for None input

        iter = 1
        new_target = []
        
        # Loop until a target value is found
        while len(new_target) == 0:
            low_pass = new_mz - (0.005 * iter)
            high_pass = new_mz + (0.005 * iter)
            iter += 1

            # Find indices of values within the low and high pass range
            matches_idx = [idx for idx, value in enumerate(mz) if low_pass < value < high_pass]
            filt_mz = mz[matches_idx]
            filt_int = intensities[matches_idx]
            
            # Get index of the maximum intensity value
            if len(filt_int) > 0:
                max_idx = [idx for idx, val in enumerate(filt_int) if val == max(filt_int)]
                if len(max_idx) > 1:
                    max_idx = max_idx[0]
                new_target = filt_mz[max_idx]

                return new_target.item()  # Return the found target m/z value

    def change_target_mz(event):
        """Update ion image with the newly selected m/z"""
        global scan_idx
        new_mz = event.xdata
        if new_mz != None:
            new_target = find_target_mz(new_mz)
            mz_entry.delete(0,tk.END)
            mz_entry.insert(0,round(new_target,4))
            plot_ion_image()
            plot_mass_spectrum(scan_idx)

    def bulk_export():
        """Export a series of ion images with the currently selected view settings. Prompts the user for a spreadsheet with target m/z and labels."""
        global fig
        #prompt the user for the target list
        target_list_file = filedialog.askopenfilename(initialdir=os.getcwd(),filetypes=[("Excel Spreadsheet",".xlsx"),("CSV File",".csv")])
        target_list = pd.read_excel(target_list_file)
        #Contextual code to clean up table depending on whether headers were included, column order, etc.
        target_list=cleanup_table(target_list,target_list_file)

        quality = qual_var.get()
        try:
            quality = float(quality)
        except:
            messagebox.showwarning(title="Invalid dpi setting...",message="dpi should be specified as a number - proceeding with default (dpi = 100)")
            quality = 100

        ##Iterate through each target m/z, drawing and then writing the ion image
        for iter,row in target_list.iterrows():
            mz_entry.delete(0,tk.END)
            mz_entry.insert(0,row.values[1])
            plot_ion_image()
            hide_patches()



            folder_name = os.path.join(os.path.dirname(file_entry.get()),"ion_images")
            img_name_base = f"{row.values[0]}-{str(row.values[1]).split(".")[0]}"
            if iter == 0:
                if os.path.exists(folder_name):
                    messagebox.showwarning(title="Folder already exists!",message="You already have an ion image folder here, please rename, move, or delete it")
                    break
                os.mkdir(folder_name)
                #Prompt user for file extension to use
                file = filedialog.asksaveasfilename(initialdir=folder_name,filetypes=[("TIF", ".tif"),("PNG",".png"),("JPG", ".jpg")],initialfile=img_name_base)
                used_extension = file.split(".")[-1]
                if used_extension not in ["tiff, tif, png, jpg"]:
                    used_extension='tif'
                    file = f"{file}.tif"
                
            else:
                file = os.path.join(folder_name,f"{img_name_base}.{used_extension}")
            
            fig.savefig(fname=file,
                        transparent=True,
                        dpi=quality,
                        format=used_extension,
                        bbox_inches="tight",
                        pad_inches=0)
        
        show_patches()

        ##optionally, export a TIC image if selected
        if include_TIC_var.get():
            view_tic_check.invoke()
            file = os.path.join(folder_name,f"TIC_Image.{used_extension}")
            fig.savefig(fname=file,
                transparent=True,
                dpi=quality,
                format=used_extension,
                bbox_inches="tight",
                pad_inches=0)


    def view_tic():
        if view_tic_option.get():
            draw_tic_image()
        else:
            plot_ion_image()        


    def draw_tic_image():
        plot_ion_image()

    def custom_NL():
        """Applies custom normalization limit as specified by the user"""
        ##0 = no custom, 1 = custom NL
        global NL_entry, norm_value, raw_ion_image

        custom_NL_desired = NL_state.get()
        if not custom_NL_desired:
            try:
                NL_entry.destroy() #remove entries for custom normalization limit when unchecked
                update_NL_button.destroy()
            except:
                pass
            update_ion_image()
        elif custom_NL_desired: ##Add the entries when checked
            norm_value = tk.StringVar(window_scout)
            if norm_value.get()=="":    
                norm_value.set(np.percentile(raw_ion_image,v_top.get()*100))
            
            NL_entry = tk.Entry(window_scout,textvariable=norm_value,highlightbackground=TEAL,bg=TEAL,background=BEIGE,fg="black",justify='center')
            NL_entry.grid(row=4,column=1)
            update_NL_button = tk.Button(window_scout,text="Get this NL",bg=TEAL,highlightbackground=TEAL,command=update_NL)
            update_NL_button.grid(row=4,column=2)
            NL_entry.bind("<Return>",update_ion_image)
            NL_entry.bind("<FocusOut>",update_ion_image)
        
        
    def update_NL():
        norm_value.set(np.percentile(raw_ion_image,v_top.get()*100))
        update_ion_image()

    def more_cmaps():
        def add_a_map():
            try:
                new_map = all_options.selection_get()
            except:
                new_map = None
            if new_map != None and not new_map in colormap_options:
                colormap_options.append(new_map)
                selected_options.insert(tk.END,new_map)

        def remove_a_map():
            try:
                remove_map = selected_options.selection_get()
            except:
                remove_map = None
            
            if remove_map != None:
                for idx, map in enumerate(colormap_options):
                    if remove_map == map:
                        del_idx = idx
                
                colormap_options.pop(del_idx)
                selected_options.delete(del_idx)

        def save_exit():
            mod_path = os.path.dirname(os.path.abspath(__file__))
            config_file = os.path.join(mod_path,"cmap_targets.json")
            json_dump = {"options":colormap_options}
            with open(config_file,'w') as file:
                json.dump(json_dump,file,indent=4)
            
            cmap_selected.set(colormap_options[0])
            cmap_selector['menu'].delete(0,'end')
            for map in colormap_options:
                cmap_selector['menu'].add_command(label=map,command=tk._setit(cmap_selected,map))
            cmap_window.destroy()

        cmap_window = tk.Tk()
        cmap_window.title("Colormap Options...")
        cmap_window.config(padx=5,pady=5,bg=TEAL)

        #Listbox for all cmap options
        avail_colmaps = tk.Label(cmap_window,text="Available Colormaps",bg=TEAL,font=FONT)
        all_options = tk.Listbox(cmap_window,bg=BEIGE,fg="black",height=10,highlightcolor=TEAL,width=35,justify='center')
        avail_colmaps.grid(row=0,column=1)
        all_options.grid(row=1,column=1,rowspan=4)

        cmaps = plt.colormaps()
        for idx, map in enumerate(cmaps):
            all_options.insert(idx,map)
        
        #Listbox for cmap options to choose from
        selected_colmaps = tk.Label(cmap_window,text="Selected Colormaps",bg=TEAL,font=FONT)
        selected_options = tk.Listbox(cmap_window,bg=BEIGE,fg="black",height=10,highlightcolor=TEAL,width=35,justify='center')
        selected_colmaps.grid(row=0,column=3)
        selected_options.grid(row=1,column=3,rowspan=4)

        mod_path = os.path.dirname(os.path.abspath(__file__))
        config_file = os.path.join(mod_path,"cmap_targets.json")
        with open(config_file,'r') as file:
            cmap_options = json.load(file)
        colormap_options = cmap_options["options"]
        for idx, map in enumerate(colormap_options):
            selected_options.insert(idx,map)

        ##Add button
        add_button=tk.Button(cmap_window,text=">>>",bg=TEAL,highlightbackground=TEAL, command=add_a_map)
        add_button.grid(row=2,column=2)

        ##Remove button
        remove_button=tk.Button(cmap_window,text="<<<",bg=TEAL,highlightbackground=TEAL, command=remove_a_map)
        remove_button.grid(row=3,column=2)

        ##Save button
        save_button=tk.Button(cmap_window,text="Save & Exit",bg=TEAL,highlightbackground=TEAL, command=save_exit)
        save_button.grid(row=5,column=2)
        
        

    ##Build the GUI window
    window_scout = tk.Tk()
    window_scout.title(f"imzML Scout v{__version__}")
    window_scout.config(padx=5,pady=5,bg=TEAL)
    style = ttk.Style()
    style.theme_use('clam')

    ##Target image:
    file_var = tk.StringVar(window_scout)
    file_var.set(tgt_file)
    file_button=tk.Button(window_scout,text="Browse for file",bg=TEAL,highlightbackground=TEAL, command=browse_for_file)
    file_entry = tk.Entry(window_scout,textvariable=file_var,highlightbackground=TEAL,background=BEIGE,fg="black",justify='center')
    file_var.trace_add('write',plot_ion_image)

    file_button.grid(row=1,column=0)
    file_entry.grid(row=1,column=1)


    ##mz entry
    mz_var = tk.StringVar(window_scout)
    mz_var.set(initial_mz)
    mz_label=tk.Label(window_scout,text="Target m/z:",bg=TEAL,font=FONT)
    mz_entry = tk.Entry(window_scout,textvariable=mz_var,highlightbackground=TEAL,background=BEIGE,fg="black",justify='center')
    mz_entry.bind("<Return>",plot_ion_image)
    mz_entry.bind("<FocusOut>",plot_ion_image)

    mz_label.grid(row=2,column=0)
    mz_entry.grid(row=2,column=1)

    ##Tolerance entry
    tol_var = tk.StringVar(window_scout)
    tol_var.set("10")
    tolerance_label=tk.Label(window_scout,text="Tolerance (ppm):",bg=TEAL,font=FONT)
    tolerance_entry = tk.Entry(window_scout,textvariable=tol_var,highlightbackground=TEAL,background=BEIGE,fg="black",justify='center')
    #tol_var.trace_add('write',plot_ion_image)
    tolerance_entry.bind("<Return>",plot_ion_image)
    tolerance_entry.bind("<FocusOut>",plot_ion_image)

    tolerance_label.grid(row=3,column=0)
    tolerance_entry.grid(row=3,column=1)

    ##Normalization buttons
    normalize_custom_entry=tk.Entry(window_scout,highlightbackground=TEAL,background=BEIGE,fg="black",justify='center')

    normalization_method = tk.StringVar(window_scout)
    no_normalization = tk.Radiobutton(window_scout,bg=TEAL,command=check_normalization,fg="white",selectcolor=TEAL,text="No Normalization",variable=normalization_method,value="none")
    no_normalization.grid(row=1,column=4)
    no_normalization.select()
    custom_method = tk.Radiobutton(window_scout,bg=TEAL,command=check_normalization,fg="white",selectcolor=TEAL,text="Custom Normalize",variable=normalization_method,value="custom")
    custom_method.grid(row=2,column=4)
    tic_method = tk.Radiobutton(window_scout,bg=TEAL,command=check_normalization,fg="white",selectcolor=TEAL,text="TIC Normalize",variable=normalization_method,value="TIC")
    tic_method.grid(row=3,column=4)

    #Smooth button
    smooth_state = tk.BooleanVar(window_scout)
    smooth_button = tk.Checkbutton(window_scout,text="Smooth Image?", bg=TEAL, font=FONT, variable=smooth_state, command=plot_ion_image)
    smooth_button.grid(row=4, column=4)

    ##Slider
    v_top = tk.DoubleVar(window_scout,value=0.96)
    v_bottom = tk.DoubleVar(window_scout,value=0.05)
    v_slider = RangeSliderV(window_scout,[v_bottom,v_top],padY=12,bgColor=TEAL,valueSide="RIGHT",font_color='#ffffff',font_family="Helvetica",line_s_color=BEIGE,digit_precision='.2f',step_size=0.01)
    v_slider.grid(row=4,column=4,rowspan=4)
    v_top.trace_add('write',update_ion_image)
    v_bottom.trace_add('write',update_ion_image)

    ##Custom normalization limit
    NL_state = tk.BooleanVar(window_scout)
    NL_checkbox = tk.Checkbutton(window_scout,text="Custom normalization limit?",bg=TEAL,font=FONT,variable=NL_state,command=custom_NL)
    NL_checkbox.grid(row=4,column=0)

    ##Colormap set
    try:
        mod_path = os.path.dirname(os.path.abspath(__file__))
        config_file = os.path.join(mod_path,"cmap_targets.json")
        with open(config_file,'r') as file:
            cmap_options = json.load(file)
        colormap_options = cmap_options["options"]
    except:
        mod_path = os.path.dirname(os.path.abspath(__file__))
        config_file = os.path.join(mod_path,"cmap_targets.json")
        json_dump = {'options':["viridis","plasma","cividis","hot","jet"]}
        with open(config_file,'w') as file:
            json.dump(json_dump,file,indent=4)

        mod_path = os.path.dirname(os.path.abspath(__file__))
        config_file = os.path.join(mod_path,"cmap_targets.json")
        with open(config_file,'r') as file:
            cmap_options = json.load(file)
        colormap_options = cmap_options["options"]
    
    cmap_selected = tk.StringVar(window_scout)
    cmap_selected.set(colormap_options[0])
    cmap_selector = tk.OptionMenu(window_scout,cmap_selected,*colormap_options)
    cmap_selector.grid(row=1,column=2)
    cmap_selected.trace_add('write',update_ion_image)

    ##cmap options
    cmap_options_button = tk.Button(window_scout,text="More cmaps...",bg=TEAL,font=TEAL,highlightbackground=TEAL,command=more_cmaps)
    cmap_options_button.grid(row=2,column=2)

    ##View TIC image
    view_tic_option = tk.BooleanVar(window_scout)
    view_tic_check = tk.Checkbutton(window_scout,text="View TIC?",bg=TEAL,font=FONT,var=view_tic_option,command=view_tic)
    view_tic_check.grid(row=3,column=2)

    ##Export buttons
    export_button=tk.Button(window_scout,text="Export Image",bg=TEAL,highlightbackground=TEAL,command=lambda:export_image(fig))
    export_button.grid(row=7,column=2)

    b_export = tk.Button(window_scout,text="Bulk Export",bg=TEAL,highlightbackground=TEAL,command=bulk_export)
    b_export.grid(row=7,column=0)

    csv_export = tk.Button(window_scout,text="csv Export",bg=TEAL,highlightbackground=TEAL,command=export_csv)
    csv_export.grid(row=8,column=2)
    b_csv_export = tk.Button(window_scout,text="Bulk csv export",bg=TEAL,highlightbackground=TEAL,command=bulk_export_csv)
    b_csv_export.grid(row=8,column=0)
    include_TIC_var = tk.BooleanVar(window_scout)
    include_tic = tk.Checkbutton(window_scout,text="Include TIC?",bg=TEAL,font=FONT,var=include_TIC_var)
    include_tic.grid(row=9,column=0)

    masked_imzml_button = tk.Button(window_scout,text="Write masked imzML", bg=TEAL, highlightbackground=TEAL,command=lambda:scout_utils.write_masked_imzml_handler(file_var.get()))
    masked_imzml_button.grid(row=9,column=3)

    ##Export quality
    quality_label = tk.Label(window_scout,text="Export quality (dpi):",bg=TEAL,font=FONT)
    qual_var = tk.StringVar(window_scout)
    qual_var.set("100")
    quality_entry = tk.Entry(window_scout,textvariable=qual_var,highlightbackground=TEAL,background=BEIGE,fg="black",justify='center')
    quality_label.grid(row=7, column=1)
    quality_entry.grid(row=8,column=1)

    #ROI Select
    load_ROI = tk.Button(window_scout, text="load ROI mask", bg=TEAL, highlightbackground=TEAL, command=scout_utils.load_ROI)
    load_ROI.grid(row=7,column=3)

    save_ROI = tk.Button(window_scout, text="save ROI mask", bg=TEAL, highlightbackground=TEAL, command=scout_utils.save_ROI)
    save_ROI.grid(row=8,column=3)

    ROI_select = tk.Button(window_scout, text="ROI Select", bg=TEAL, highlightbackground=TEAL, command=lambda:scout_utils.ROI_select(ROI_select,ion_image,aspect_ratio,color_NL))
    ROI_select.grid(row=7, column = 4)

    apply_ROI_mask = tk.BooleanVar(window_scout)
    apply_ROI = tk.Checkbutton(window_scout, text="Apply ROI mask?", bg=TEAL, font=FONT, var=apply_ROI_mask, command=update_ion_image)
    apply_ROI.grid(row=8, column=4)



    start_var = tk.StringVar(window_scout)
    start_var.set(None)
    end_var = tk.StringVar(window_scout)
    end_var.set(None)
    first_img = tk.BooleanVar(window_scout)
    first_img.set(True)
    first_MS = tk.BooleanVar(window_scout)
    first_MS.set(True)
    on_startup = True

    if on_startup:
        if tgt_file != "":
            plot_ion_image()
        
        on_startup=False

    window_scout.mainloop()

if __name__ == "__main__":
    try:
        main(sys.argv[1])
    except:
        main()
